/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _GitHubActionsCache_instances, _a, _GitHubActionsCache_compressionMethodPromise, _GitHubActionsCache_compressionMethod_get, _GitHubActionsCache_baseUrl, _GitHubActionsCache_authToken, _GitHubActionsCache_logger, _GitHubActionsCache_emptyDirectoriesManifestPath, _GitHubActionsCache_computeCacheKey, _GitHubActionsCache_computeVersion, _GitHubActionsCache_makeAuthenticatedHttpClient, _GitHubActionsCache_makeTarball, _GitHubActionsCache_checkForCacheEntry, _GitHubActionsCache_reserveCacheEntry, _GitHubActionsCacheHit_instances, _GitHubActionsCacheHit_url, _GitHubActionsCacheHit_compressionMethod, _GitHubActionsCacheHit_applied, _GitHubActionsCacheHit_emptyDirectoriesManifestPath, _GitHubActionsCacheHit_createEmptyDirectories;
import * as pathlib from 'path';
import * as fs from 'fs/promises';
import { createHash } from 'crypto';
import * as cacheUtils from '@actions/cache/lib/internal/cacheUtils.js';
import { createTar, extractTar } from '@actions/cache/lib/internal/tar.js';
import { retryTypedResponse } from '@actions/cache/lib/internal/requestUtils.js';
import { saveCache, downloadCache, } from '@actions/cache/lib/internal/cacheHttpClient.js';
import { HttpClient } from '@actions/http-client';
import { BearerCredentialHandler } from '@actions/http-client/auth.js';
import { isSuccessStatusCode } from '@actions/cache/lib/internal/requestUtils.js';
import { scriptReferenceToString } from '../script.js';
import { getScriptDataDir } from '../util/script-data-dir.js';
// TODO(aomarks) Consider dropping the dependency on @actions/cache by writing
// our own implementation. See https://github.com/google/wireit/issues/107 for
// details.
//
// The public API provided by the @actions/cache package doesn't exactly meet
// our needs, because it automatically uses the file paths that are included in
// the tarball as part of the cache entry version (see
// https://github.com/actions/toolkit/blob/7654d97eb6c4a3d564f036a2d4a783ae9105ec07/packages/cache/src/internal/cacheHttpClient.ts#L70),
// and implements globbing differently.
//
// We want complete control over our cache key, instead of having it be
// generated automatically based on file paths.
//
// For this reason, we are reaching into the "internal/" directory of this
// package to get more control. This is bad because those modules could change
// at any time, which is why we currently have a strict ("=") version pin in our
// package.json.
//
// The @actions/cache package is also our largest dependency by far. It's 22MB,
// and adds 63 transitive dependencies.
//
// The logic in here could be re-implemented in a fairly minimal way. The main
// tricky part is the way it handles tarball generation across platforms
// (https://github.com/actions/toolkit/blob/7654d97eb6c4a3d564f036a2d4a783ae9105ec07/packages/cache/src/internal/tar.ts).
/**
 * Caches script output to the GitHub Actions caching service.
 */
export class GitHubActionsCache {
    constructor(logger) {
        _GitHubActionsCache_instances.add(this);
        _GitHubActionsCache_baseUrl.set(this, void 0);
        _GitHubActionsCache_authToken.set(this, void 0);
        _GitHubActionsCache_logger.set(this, void 0);
        // The ACTIONS_CACHE_URL and ACTIONS_RUNTIME_TOKEN environment variables are
        // automatically provided to GitHub Actions re-usable workflows. However,
        // they are _not_ provided to regular "run" scripts. For this reason, we
        // re-export those variables so that all "run" scripts can access them using
        // the "google/wireit@setup-github-actions-caching/v1" re-usable workflow.
        const baseUrl = process.env['ACTIONS_CACHE_URL'];
        if (!baseUrl) {
            throw new GitHubActionsCacheError('invalid-usage', 'The ACTIONS_CACHE_URL variable was not set, but is required when ' +
                'WIREIT_CACHE=github. Use the google/wireit@setup-github-cache/v1 ' +
                'action to automatically set environment variables.');
        }
        if (!baseUrl.endsWith('/')) {
            // Internally, the @actions/cache library expects the URL to end with a
            // slash. While we could be more lenient, we want to match the behavior of
            // any other calls happening inside that library which we don't control.
            throw new GitHubActionsCacheError('invalid-usage', `The ACTIONS_CACHE_URL must end in a forward-slash, got ${JSON.stringify(baseUrl)}.`);
        }
        __classPrivateFieldSet(this, _GitHubActionsCache_baseUrl, baseUrl, "f");
        const authToken = process.env['ACTIONS_RUNTIME_TOKEN'];
        if (!authToken) {
            throw new GitHubActionsCacheError('invalid-usage', 'The ACTIONS_RUNTIME_TOKEN variable was not set, but is required when ' +
                'WIREIT_CACHE=github. Use the google/wireit@setup-github-cache/v1 ' +
                'action to automatically set environment variables.');
        }
        __classPrivateFieldSet(this, _GitHubActionsCache_authToken, authToken, "f");
        __classPrivateFieldSet(this, _GitHubActionsCache_logger, logger, "f");
    }
    async get(script, stateStr) {
        const compressionMethod = await __classPrivateFieldGet(GitHubActionsCache, _a, "a", _GitHubActionsCache_compressionMethod_get);
        const version = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeVersion).call(this, stateStr, compressionMethod);
        const location = await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_checkForCacheEntry).call(this, __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeCacheKey).call(this, script), version);
        if (location === undefined) {
            // No cache hit.
            return undefined;
        }
        return new GitHubActionsCacheHit(location, compressionMethod, __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_emptyDirectoriesManifestPath).call(this, script, version));
    }
    async set(script, stateStr, relativeFiles) {
        // We're going to build a tarball. We do this by passing paths to the "tar"
        // command. When we pass a directory to "tar", all of its children are
        // implicitly included. This is a problem, because sometimes we want to add
        // a directory, but only a subset of its children (or none of them), because
        // of exclusion patterns in the "output" globs.
        //
        // To work around this problem, we never pass directories to "tar". Instead,
        // we enumerate every specific file. This works fine, except for empty
        // directories. If a directory was explicitly listed for inclusion, and it
        // happens to be empty, we should cache an empty directory.
        //
        // So we need special handling for empty directories. If there are empty
        // directories, we create a special "empty directories manifest" file, and
        // include that in the tarball at a predictable location (inside the
        // ".wireit" folder so that it won't collide with a user file). Then when we
        // untar, we check for that manifest, create any missing directories listed
        // inside it, and delete the manifest.
        const files = new Set();
        const emptyDirs = new Set();
        {
            const nonEmptyDirs = new Set();
            for (const { path, dirent } of relativeFiles) {
                const absPath = pathlib.join(script.packageDir, path);
                if (dirent.isDirectory()) {
                    // Initially assume every directory might be empty. We might see a
                    // directory entry after a file that's in it. We'll filter it down
                    // after this loop.
                    emptyDirs.add(absPath);
                }
                else {
                    files.add(absPath);
                    // Add all parent directories of this file to set of non-empty
                    // directories.
                    let cur = pathlib.dirname(absPath);
                    while (!nonEmptyDirs.has(cur)) {
                        // Note if we've already added the child, we must have already added
                        // all of its parents too.
                        nonEmptyDirs.add(cur);
                        cur = pathlib.dirname(cur);
                    }
                }
            }
            for (const nonEmptyDir of nonEmptyDirs) {
                emptyDirs.delete(nonEmptyDir);
            }
        }
        const compressionMethod = await __classPrivateFieldGet(GitHubActionsCache, _a, "a", _GitHubActionsCache_compressionMethod_get);
        const version = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeVersion).call(this, stateStr, compressionMethod);
        let emptyDirsManifestPath;
        if (emptyDirs.size > 0) {
            const emptyDirsManifest = JSON.stringify([...emptyDirs]);
            emptyDirsManifestPath = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_emptyDirectoriesManifestPath).call(this, script, version);
            await fs.mkdir(pathlib.dirname(emptyDirsManifestPath), { recursive: true });
            await fs.writeFile(emptyDirsManifestPath, emptyDirsManifest, {
                encoding: 'utf8',
            });
            files.add(emptyDirsManifestPath);
        }
        const tarballPath = await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_makeTarball).call(this, [...files], compressionMethod);
        try {
            const tarBytes = cacheUtils.getArchiveFileSizeInBytes(tarballPath);
            // Reference: https://github.com/actions/toolkit/blob/f8a69bc473af4a204d0c03de61d5c9d1300dfb17/packages/cache/src/cache.ts#L174
            const GB = 1024 * 1024 * 1024;
            const maxBytes = 10 * GB;
            if (tarBytes > maxBytes) {
                __classPrivateFieldGet(this, _GitHubActionsCache_logger, "f").log({
                    script,
                    type: 'info',
                    detail: 'generic',
                    message: `Output was too big to be cached: ` +
                        `${Math.round(tarBytes / GB)}GB > ` +
                        `${Math.round(maxBytes / GB)}GB.`,
                });
                return false;
            }
            const id = await __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_reserveCacheEntry).call(this, __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_makeAuthenticatedHttpClient).call(this), __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_computeCacheKey).call(this, script), version, tarBytes);
            // It's likely that we'll occasionally fail to reserve an entry and get
            // undefined here, especially when running multiple GitHub Action jobs in
            // parallel with the same scripts, because there is a window of time
            // between calling "get" and "set" on the cache in which another worker
            // could have reserved the entry before us. Non fatal, just don't save.
            if (id !== undefined) {
                await saveCache(id, tarballPath);
            }
        }
        finally {
            // Delete the tarball.
            const tarballDeleted = cacheUtils.unlinkFile(tarballPath);
            // Also delete the empty directories manifest file.
            if (emptyDirsManifestPath !== undefined) {
                await fs.unlink(emptyDirsManifestPath);
            }
            await tarballDeleted;
        }
        return true;
    }
}
_a = GitHubActionsCache, _GitHubActionsCache_baseUrl = new WeakMap(), _GitHubActionsCache_authToken = new WeakMap(), _GitHubActionsCache_logger = new WeakMap(), _GitHubActionsCache_instances = new WeakSet(), _GitHubActionsCache_compressionMethod_get = function _GitHubActionsCache_compressionMethod_get() {
    // We only need to do this once per Wireit process. It's expensive because
    // it spawns a child process to test the tar version.
    if (__classPrivateFieldGet(this, _a, "f", _GitHubActionsCache_compressionMethodPromise) === undefined) {
        __classPrivateFieldSet(this, _a, cacheUtils.getCompressionMethod(), "f", _GitHubActionsCache_compressionMethodPromise);
    }
    return __classPrivateFieldGet(this, _a, "f", _GitHubActionsCache_compressionMethodPromise);
}, _GitHubActionsCache_emptyDirectoriesManifestPath = function _GitHubActionsCache_emptyDirectoriesManifestPath(script, version) {
    return pathlib.join(getScriptDataDir(script), `github-cache-empty-directories-manifest-${version}.json`);
}, _GitHubActionsCache_computeCacheKey = function _GitHubActionsCache_computeCacheKey(script) {
    return createHash('sha256')
        .update(scriptReferenceToString(script))
        .digest('hex');
}, _GitHubActionsCache_computeVersion = function _GitHubActionsCache_computeVersion(stateStr, compressionMethod) {
    return createHash('sha256')
        .update([
        stateStr,
        compressionMethod,
        // The ImageOS environment variable tells us which operating system
        // version is being used for the worker VM (e.g. "ubuntu20",
        // "macos11"). We already include process.platform in ScriptState, but
        // this is more specific.
        //
        // There is also an ImageVersion variable (e.g. "20220405.4") which we
        // could consider including, but it probably changes frequently and is
        // unlikely to affect output, so we prefer the higher cache hit rate.
        process.env.ImageOS ?? '',
    ].join('\x1E') // ASCII record seperator
    )
        .digest('hex');
}, _GitHubActionsCache_makeAuthenticatedHttpClient = function _GitHubActionsCache_makeAuthenticatedHttpClient() {
    const bearerCredentialHandler = new BearerCredentialHandler(__classPrivateFieldGet(this, _GitHubActionsCache_authToken, "f"));
    return new HttpClient('actions/cache', [bearerCredentialHandler], {
        headers: {
            Accept: 'application/json;api-version=6.0-preview.1',
        },
    });
}, _GitHubActionsCache_makeTarball = 
/**
 * Create a tarball file in a local temp directory containing the given paths.
 *
 * @returns The full path to the tarball file on disk.
 */
async function _GitHubActionsCache_makeTarball(paths, compressionMethod) {
    const folder = await cacheUtils.createTempDirectory();
    await createTar(folder, paths, compressionMethod);
    const path = pathlib.join(folder, cacheUtils.getCacheFileName(compressionMethod));
    return path;
}, _GitHubActionsCache_checkForCacheEntry = 
/**
 * Check for a cache entry.
 *
 * @returns A tarball URL if this cache entry exists, or undefined if it does
 * not exist.
 */
async function _GitHubActionsCache_checkForCacheEntry(key, version) {
    const httpClient = __classPrivateFieldGet(this, _GitHubActionsCache_instances, "m", _GitHubActionsCache_makeAuthenticatedHttpClient).call(this);
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
    const response = await retryTypedResponse('getCacheEntry', async () => httpClient.getJson(
    /** For docs on this API, see {@link FakeGitHubActionsCacheServer} */
    `${__classPrivateFieldGet(this, _GitHubActionsCache_baseUrl, "f")}_apis/artifactcache/cache?keys=${encodeURIComponent(key)}&version=${version}`));
    if (response.statusCode === /* No Content */ 204) {
        return undefined;
    }
    if (!isSuccessStatusCode(response.statusCode) ||
        response.error !== undefined ||
        response.result?.archiveLocation === undefined) {
        throw new Error(`Error getting cache entry: ${response.statusCode} ${response.error?.message ?? '<no error message>'}`);
    }
    return response.result.archiveLocation;
}, _GitHubActionsCache_reserveCacheEntry = 
/**
 * Reserve a cache entry.
 *
 * @returns A numeric cache id the cache entry was reserved for us, or
 * undefined if the cache entry was already reserved.
 */
async function _GitHubActionsCache_reserveCacheEntry(httpClient, key, version, cacheSize) {
    const request = {
        key,
        version,
        cacheSize,
    };
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
    const response = await retryTypedResponse('reserveCache', async () => httpClient.postJson(
    /** For docs on this API, see {@link FakeGitHubActionsCacheServer} */
    `${__classPrivateFieldGet(this, _GitHubActionsCache_baseUrl, "f")}_apis/artifactcache/caches`, request));
    if (response.statusCode === /* Conflict */ 409) {
        // This cache entry has already been reserved, so we can't write to it.
        return undefined;
    }
    if (!isSuccessStatusCode(response.statusCode) ||
        response.error !== undefined ||
        response.result?.cacheId === undefined) {
        throw new Error(`Error reserving cache entry: ${response.statusCode} ${response.error?.message ?? '<no error message>'}`);
    }
    return response.result.cacheId;
};
_GitHubActionsCache_compressionMethodPromise = { value: void 0 };
class GitHubActionsCacheHit {
    constructor(location, compressionMethod, emptyDirectoriesManifestPath) {
        _GitHubActionsCacheHit_instances.add(this);
        _GitHubActionsCacheHit_url.set(this, void 0);
        _GitHubActionsCacheHit_compressionMethod.set(this, void 0);
        _GitHubActionsCacheHit_applied.set(this, false);
        _GitHubActionsCacheHit_emptyDirectoriesManifestPath.set(this, void 0);
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_url, location, "f");
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_compressionMethod, compressionMethod, "f");
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_emptyDirectoriesManifestPath, emptyDirectoriesManifestPath, "f");
    }
    async apply() {
        if (__classPrivateFieldGet(this, _GitHubActionsCacheHit_applied, "f")) {
            throw new Error('GitHubActionsCacheHit.apply was called more than once');
        }
        __classPrivateFieldSet(this, _GitHubActionsCacheHit_applied, true, "f");
        const archivePath = pathlib.join(await cacheUtils.createTempDirectory(), cacheUtils.getCacheFileName(__classPrivateFieldGet(this, _GitHubActionsCacheHit_compressionMethod, "f")));
        try {
            await downloadCache(__classPrivateFieldGet(this, _GitHubActionsCacheHit_url, "f"), archivePath);
            await extractTar(archivePath, __classPrivateFieldGet(this, _GitHubActionsCacheHit_compressionMethod, "f"));
            await __classPrivateFieldGet(this, _GitHubActionsCacheHit_instances, "m", _GitHubActionsCacheHit_createEmptyDirectories).call(this);
        }
        finally {
            await cacheUtils.unlinkFile(archivePath);
        }
    }
}
_GitHubActionsCacheHit_url = new WeakMap(), _GitHubActionsCacheHit_compressionMethod = new WeakMap(), _GitHubActionsCacheHit_applied = new WeakMap(), _GitHubActionsCacheHit_emptyDirectoriesManifestPath = new WeakMap(), _GitHubActionsCacheHit_instances = new WeakSet(), _GitHubActionsCacheHit_createEmptyDirectories = async function _GitHubActionsCacheHit_createEmptyDirectories() {
    let manifest;
    try {
        manifest = await fs.readFile(__classPrivateFieldGet(this, _GitHubActionsCacheHit_emptyDirectoriesManifestPath, "f"), {
            encoding: 'utf8',
        });
    }
    catch (error) {
        const { code } = error;
        if (code === 'ENOENT') {
            // No empty dirs manifest means no empty dirs which is no problem.
            return;
        }
        throw error;
    }
    const emptyDirs = JSON.parse(manifest);
    await Promise.all([
        ...emptyDirs.map((dir) => fs.mkdir(dir, { recursive: true })),
        fs.unlink(__classPrivateFieldGet(this, _GitHubActionsCacheHit_emptyDirectoriesManifestPath, "f")),
    ]);
};
/**
 * An exception thrown by {@link GitHubActionsCache}.
 *
 * Note we don't use {@link WireitError} here because we don't have the full
 * context of the script we're trying to evaluate.
 */
export class GitHubActionsCacheError extends Error {
    constructor(reason, message) {
        super(message);
        Object.defineProperty(this, "reason", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.reason = reason;
    }
}
//# sourceMappingURL=github-actions-cache.js.map