/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Executor_executions, _Executor_logger, _Executor_workerPool, _Executor_cache, _ScriptExecution_instances, _ScriptExecution_script, _ScriptExecution_executor, _ScriptExecution_cache, _ScriptExecution_workerPool, _ScriptExecution_logger, _ScriptExecution_execute, _ScriptExecution_anyInputFilesDeletedSinceLastRun, _ScriptExecution_executeDependencies, _ScriptExecution_executeCommandIfNeeded, _ScriptExecution_pathEnvironmentVariable_get, _ScriptExecution_saveToCacheIfPossible, _ScriptExecution_computeState, _ScriptExecution_dataDir_get, _ScriptExecution_statePath_get, _ScriptExecution_stdoutReplayPath_get, _ScriptExecution_stderrReplayPath_get, _ScriptExecution_readPreviousState, _ScriptExecution_writeStateFile, _ScriptExecution_prepareDataDir, _ScriptExecution_cleanOutput, _ScriptExecution_replayStdoutIfPresent, _ScriptExecution_replayStderrIfPresent;
import * as pathlib from 'path';
import * as fs from 'fs/promises';
import { createHash } from 'crypto';
import { createReadStream, createWriteStream } from 'fs';
import { spawn } from 'child_process';
import { WireitError } from './error.js';
import { scriptReferenceToString } from './script.js';
import { shuffle } from './util/shuffle.js';
import { getScriptDataDir } from './util/script-data-dir.js';
import { unreachable } from './util/unreachable.js';
import { glob } from './util/glob.js';
import { deleteEntries } from './util/delete.js';
import { AggregateError } from './util/aggregate-error.js';
import { augmentProcessEnvSafelyIfOnWindows, posixifyPathIfOnWindows, } from './util/windows.js';
/**
 * The PATH environment variable of this process, minus all of the leading
 * "node_modules/.bin" entries that the incoming "npm run" command already set.
 *
 * We want full control over which "node_modules/.bin" paths are in the PATH of
 * the processes we spawn, so that cross-package dependencies act as though we
 * are running "npm run" with each package as the cwd.
 *
 * We only need to do this once per Wireit process, because process.env never
 * changes.
 */
const PATH_ENV_SUFFIX = (() => {
    const path = process.env.PATH ?? '';
    // Note the PATH delimiter is platform-dependent.
    const entries = path.split(pathlib.delimiter);
    const nodeModulesBinSuffix = pathlib.join('node_modules', '.bin');
    const endOfNodeModuleBins = entries.findIndex((entry) => !entry.endsWith(nodeModulesBinSuffix));
    return entries.slice(endOfNodeModuleBins).join(pathlib.delimiter);
})();
/**
 * Executes a script that has been analyzed and validated by the Analyzer.
 */
export class Executor {
    constructor(logger, workerPool, cache) {
        _Executor_executions.set(this, new Map());
        _Executor_logger.set(this, void 0);
        _Executor_workerPool.set(this, void 0);
        _Executor_cache.set(this, void 0);
        __classPrivateFieldSet(this, _Executor_logger, logger, "f");
        __classPrivateFieldSet(this, _Executor_workerPool, workerPool, "f");
        __classPrivateFieldSet(this, _Executor_cache, cache, "f");
    }
    async execute(script) {
        const executionKey = scriptReferenceToString(script);
        let promise = __classPrivateFieldGet(this, _Executor_executions, "f").get(executionKey);
        if (promise === undefined) {
            promise = ScriptExecution.execute(script, this, __classPrivateFieldGet(this, _Executor_workerPool, "f"), __classPrivateFieldGet(this, _Executor_cache, "f"), __classPrivateFieldGet(this, _Executor_logger, "f"));
            __classPrivateFieldGet(this, _Executor_executions, "f").set(executionKey, promise);
        }
        return promise;
    }
}
_Executor_executions = new WeakMap(), _Executor_logger = new WeakMap(), _Executor_workerPool = new WeakMap(), _Executor_cache = new WeakMap();
/**
 * A single execution of a specific script.
 */
class ScriptExecution {
    constructor(script, executor, workerPool, cache, logger) {
        _ScriptExecution_instances.add(this);
        _ScriptExecution_script.set(this, void 0);
        _ScriptExecution_executor.set(this, void 0);
        _ScriptExecution_cache.set(this, void 0);
        _ScriptExecution_workerPool.set(this, void 0);
        _ScriptExecution_logger.set(this, void 0);
        __classPrivateFieldSet(this, _ScriptExecution_script, script, "f");
        __classPrivateFieldSet(this, _ScriptExecution_executor, executor, "f");
        __classPrivateFieldSet(this, _ScriptExecution_workerPool, workerPool, "f");
        __classPrivateFieldSet(this, _ScriptExecution_cache, cache, "f");
        __classPrivateFieldSet(this, _ScriptExecution_logger, logger, "f");
    }
    static execute(script, executor, workerPool, cache, logger) {
        var _a;
        return __classPrivateFieldGet((_a = new ScriptExecution(script, executor, workerPool, cache, logger)), _ScriptExecution_instances, "m", _ScriptExecution_execute).call(_a);
    }
}
_ScriptExecution_script = new WeakMap(), _ScriptExecution_executor = new WeakMap(), _ScriptExecution_cache = new WeakMap(), _ScriptExecution_workerPool = new WeakMap(), _ScriptExecution_logger = new WeakMap(), _ScriptExecution_instances = new WeakSet(), _ScriptExecution_execute = async function _ScriptExecution_execute() {
    const dependencyStates = await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_executeDependencies).call(this);
    // Note we must wait for dependencies to finish before generating the cache
    // key, because a dependency could create or modify an input file to this
    // script, which would affect the key.
    const state = await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_computeState).call(this, dependencyStates);
    const stateStr = JSON.stringify(state);
    const prevStateStr = await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_readPreviousState).call(this);
    if (state.cacheable &&
        prevStateStr !== undefined &&
        prevStateStr === stateStr) {
        // TODO(aomarks) Does not preserve original order of stdout vs stderr
        // chunks. See https://github.com/google/wireit/issues/74.
        await Promise.all([
            __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_replayStdoutIfPresent).call(this),
            __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_replayStderrIfPresent).call(this),
        ]);
        __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
            script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
            type: 'success',
            reason: 'fresh',
        });
        return state;
    }
    // It's important that we delete any previous state before running the
    // command or restoring from cache, because if either fails mid-flight, we
    // don't want to think that the previous state is still valid.
    await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_prepareDataDir).call(this);
    const cacheHit = state.cacheable
        ? await __classPrivateFieldGet(this, _ScriptExecution_cache, "f")?.get(__classPrivateFieldGet(this, _ScriptExecution_script, "f"), stateStr)
        : undefined;
    const shouldClean = (() => {
        if (cacheHit !== undefined) {
            // If we are restoring from cache, we should always delete existing
            // output. The purpose of "clean:false" and "clean:if-file-deleted" is to
            // allow tools with incremental build (like tsc --build) to work.
            //
            // However, this only applies when the tool is able to observe each
            // incremental change to the input files. When we restore from cache, we
            // are directly replacing the output files, and not invoking the tool at
            // all, so there is no way for the tool to do any cleanup.
            return true;
        }
        switch (__classPrivateFieldGet(this, _ScriptExecution_script, "f").clean) {
            case true: {
                return true;
            }
            case false: {
                return false;
            }
            case 'if-file-deleted': {
                if (prevStateStr === undefined) {
                    // If we don't know the previous state, then we can't know whether
                    // any input files were removed. It's safer to err on the side of
                    // cleaning.
                    return true;
                }
                return __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_anyInputFilesDeletedSinceLastRun).call(this, state, JSON.parse(prevStateStr));
            }
            default: {
                throw new Error(`Unhandled clean setting: ${unreachable(__classPrivateFieldGet(this, _ScriptExecution_script, "f").clean)}`);
            }
        }
    })();
    if (shouldClean) {
        await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_cleanOutput).call(this);
    }
    if (cacheHit !== undefined) {
        await cacheHit.apply();
        // We include stdout and stderr replay files when we save to the cache, so
        // if there were any, they will now be in place.
        // TODO(aomarks) Does not preserve original order of stdout vs stderr
        // chunks. See https://github.com/google/wireit/issues/74.
        await Promise.all([
            __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_replayStdoutIfPresent).call(this),
            __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_replayStderrIfPresent).call(this),
        ]);
        __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
            script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
            type: 'success',
            reason: 'cached',
        });
    }
    else {
        await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_executeCommandIfNeeded).call(this);
    }
    // TODO(aomarks) We don't technically need to wait for these to finish to
    // return, we only need to wait in the top-level call to execute. The same
    // will go for saving output to the cache.
    await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_writeStateFile).call(this, stateStr);
    if (cacheHit === undefined && state.cacheable) {
        await __classPrivateFieldGet(this, _ScriptExecution_instances, "m", _ScriptExecution_saveToCacheIfPossible).call(this, stateStr);
    }
    return state;
}, _ScriptExecution_anyInputFilesDeletedSinceLastRun = function _ScriptExecution_anyInputFilesDeletedSinceLastRun(curState, prevState) {
    const curFiles = Object.keys(curState.files);
    const prevFiles = Object.keys(prevState.files);
    if (curFiles.length < prevFiles.length) {
        return true;
    }
    const newFilesSet = new Set(curFiles);
    for (const oldFile of prevFiles) {
        if (!newFilesSet.has(oldFile)) {
            return true;
        }
    }
    return false;
}, _ScriptExecution_executeDependencies = async function _ScriptExecution_executeDependencies() {
    // Randomize the order we execute dependencies to make it less likely for a
    // user to inadvertently depend on any specific order, which could indicate
    // a missing edge in the dependency graph.
    shuffle(__classPrivateFieldGet(this, _ScriptExecution_script, "f").dependencies);
    // Note we use Promise.allSettled instead of Promise.all so that we can
    // collect all errors, instead of just the first one.
    const dependencyResults = await Promise.allSettled(__classPrivateFieldGet(this, _ScriptExecution_script, "f").dependencies.map((dependency) => __classPrivateFieldGet(this, _ScriptExecution_executor, "f").execute(dependency)));
    const errors = [];
    const results = [];
    for (let i = 0; i < dependencyResults.length; i++) {
        const result = dependencyResults[i];
        if (result.status === 'rejected') {
            const error = result.reason;
            if (error instanceof AggregateError) {
                // Flatten nested AggregateErrors.
                errors.push(...error.errors);
            }
            else {
                errors.push(error);
            }
        }
        else {
            results.push([__classPrivateFieldGet(this, _ScriptExecution_script, "f").dependencies[i], result.value]);
        }
    }
    if (errors.length > 0) {
        throw errors.length === 1 ? errors[0] : new AggregateError(errors);
    }
    return results;
}, _ScriptExecution_executeCommandIfNeeded = async function _ScriptExecution_executeCommandIfNeeded() {
    // It's valid to not have a command defined, since thats a useful way to
    // alias a group of dependency scripts. In this case, we can return early.
    if (!__classPrivateFieldGet(this, _ScriptExecution_script, "f").command) {
        __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
            script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
            type: 'success',
            reason: 'no-command',
        });
        return;
    }
    __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
        script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
        type: 'info',
        detail: 'running',
    });
    const command = __classPrivateFieldGet(this, _ScriptExecution_script, "f").command;
    await __classPrivateFieldGet(this, _ScriptExecution_workerPool, "f").run(async () => {
        // TODO(aomarks) Update npm_ environment variables to reflect the new
        // package.
        const child = spawn(command, {
            cwd: __classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir,
            // Conveniently, "shell:true" has the same shell-selection behavior as
            // "npm run", where on macOS and Linux it is "sh", and on Windows it is
            // %COMSPEC% || "cmd.exe".
            //
            // References:
            //   https://nodejs.org/api/child_process.html#child_processspawncommand-args-options
            //   https://nodejs.org/api/child_process.html#default-windows-shell
            //   https://github.com/npm/run-script/blob/a5b03bdfc3a499bf7587d7414d5ea712888bfe93/lib/make-spawn-args.js#L11
            shell: true,
            env: augmentProcessEnvSafelyIfOnWindows({
                PATH: __classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_pathEnvironmentVariable_get),
            }),
        });
        // Only create the stdout/stderr replay files if we encounter anything on
        // this streams.
        let stdoutReplay;
        let stderrReplay;
        child.stdout.on('data', (data) => {
            __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
                script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                type: 'output',
                stream: 'stdout',
                data,
            });
            if (stdoutReplay === undefined) {
                stdoutReplay = createWriteStream(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stdoutReplayPath_get));
            }
            stdoutReplay.write(data);
        });
        child.stderr.on('data', (data) => {
            __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
                script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                type: 'output',
                stream: 'stderr',
                data,
            });
            if (stderrReplay === undefined) {
                stderrReplay = createWriteStream(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stderrReplayPath_get));
            }
            stderrReplay.write(data);
        });
        const completed = new Promise((resolve, reject) => {
            child.on('error', (error) => {
                reject(new WireitError({
                    script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                    type: 'failure',
                    reason: 'spawn-error',
                    message: error.message,
                }));
            });
            child.on('close', (status, signal) => {
                if (signal !== null) {
                    reject(new WireitError({
                        script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                        type: 'failure',
                        reason: 'signal',
                        signal,
                    }));
                }
                else if (status !== 0) {
                    reject(new WireitError({
                        script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                        type: 'failure',
                        reason: 'exit-non-zero',
                        // status should only ever be null if signal was not null, but
                        // this isn't reflected in the TypeScript types. Just in case, and
                        // to make TypeScript happy, fall back to -1 (which is a
                        // conventional exit status used for "exited with signal").
                        status: status ?? -1,
                    }));
                }
                else {
                    resolve();
                }
            });
        });
        try {
            await completed;
        }
        finally {
            if (stdoutReplay !== undefined) {
                await closeWriteStream(stdoutReplay);
            }
            if (stderrReplay !== undefined) {
                await closeWriteStream(stderrReplay);
            }
        }
    });
    __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
        script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
        type: 'success',
        reason: 'exit-zero',
    });
}, _ScriptExecution_pathEnvironmentVariable_get = function _ScriptExecution_pathEnvironmentVariable_get() {
    // Given package "/foo/bar", walk up the path hierarchy to generate
    // "/foo/bar/node_modules/.bin:/foo/node_modules/.bin:/node_modules/.bin".
    const entries = [];
    let cur = __classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir;
    while (true) {
        entries.push(pathlib.join(cur, 'node_modules', '.bin'));
        const parent = pathlib.dirname(cur);
        if (parent === cur) {
            break;
        }
        cur = parent;
    }
    // Add the inherited PATH variable, minus any "node_modules/.bin" entries
    // that were set by the "npm run" command that spawned Wireit.
    entries.push(PATH_ENV_SUFFIX);
    // Note the PATH delimiter is platform-dependent.
    return entries.join(pathlib.delimiter);
}, _ScriptExecution_saveToCacheIfPossible = 
/**
 * Save the current output files to the configured cache if possible.
 */
async function _ScriptExecution_saveToCacheIfPossible(stateStr) {
    if (__classPrivateFieldGet(this, _ScriptExecution_cache, "f") === undefined || __classPrivateFieldGet(this, _ScriptExecution_script, "f").output === undefined) {
        return;
    }
    await __classPrivateFieldGet(this, _ScriptExecution_cache, "f").set(__classPrivateFieldGet(this, _ScriptExecution_script, "f"), stateStr, await glob([
        ...__classPrivateFieldGet(this, _ScriptExecution_script, "f").output,
        // Also include the "stdout" and "stderr" replay files at their
        // standard location within the ".wireit" directory for this script so
        // that we can replay them after restoring.
        //
        // We're passing this to #glob because it's an easy way to only
        // include them only if they exist. We don't want to include files
        // that don't exist becuase then we'll make empty directories and will
        // get an error from fs.cp.
        //
        // Convert to relative paths because we want to pass relative paths to
        // Cache.set, but fast-glob doesn't automatically relativize to the
        // cwd when passing an absolute path.
        //
        // Convert Windows-style paths to POSIX-style paths if we are on
        // Windows, because fast-glob only understands POSIX-style paths.
        posixifyPathIfOnWindows(pathlib.relative(__classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir, __classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stdoutReplayPath_get))),
        posixifyPathIfOnWindows(pathlib.relative(__classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir, __classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stderrReplayPath_get))),
    ], {
        cwd: __classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir,
        absolute: false,
        followSymlinks: false,
        includeDirectories: true,
        expandDirectories: true,
    }));
}, _ScriptExecution_computeState = 
/**
 * Generate the state object for this script based on its current input files,
 * and the state of its dependencies.
 */
async function _ScriptExecution_computeState(dependencyStates) {
    let allDependenciesAreCacheable = true;
    const filteredDependencyStates = [];
    for (const [dep, depState] of dependencyStates) {
        if (!depState.cacheable) {
            allDependenciesAreCacheable = false;
        }
        filteredDependencyStates.push([scriptReferenceToString(dep), depState]);
    }
    let fileHashes;
    if (__classPrivateFieldGet(this, _ScriptExecution_script, "f").files?.length) {
        const files = await glob(__classPrivateFieldGet(this, _ScriptExecution_script, "f").files, {
            cwd: __classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir,
            absolute: false,
            followSymlinks: true,
            // TODO(aomarks) This means that empty directories are not reflected in
            // the state, however an empty directory could modify the behavior of a
            // script. We should probably include empty directories; we'll just need
            // special handling when we compute the state key, because there is no
            // hash we can compute.
            includeDirectories: false,
            // We must expand directories here, because we need the complete
            // explicit list of files to hash.
            expandDirectories: true,
        });
        // TODO(aomarks) Instead of reading and hashing every input file on every
        // build, use inode/mtime/ctime/size metadata (which is much faster to
        // read) as a heuristic to detect files that have likely changed, and
        // otherwise re-use cached hashes that we store in e.g.
        // ".wireit/<script>/hashes".
        fileHashes = await Promise.all(files.map(async (file) => {
            const absolutePath = pathlib.resolve(__classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir, file.path);
            const hash = createHash('sha256');
            for await (const chunk of createReadStream(absolutePath)) {
                hash.update(chunk);
            }
            return [file.path, hash.digest('hex')];
        }));
    }
    else {
        fileHashes = [];
    }
    const cacheable = 
    // If command is undefined, then it's always safe to be cached, because
    // the script isn't going to do anything anyway. In these cases, the state
    // is essentially just the state of the dependencies.
    __classPrivateFieldGet(this, _ScriptExecution_script, "f").command === undefined ||
        // Otherwise, If files are undefined, then it's not safe to be cached,
        // because we don't know what the inputs are, so we can't know if the
        // output of this script could change.
        (__classPrivateFieldGet(this, _ScriptExecution_script, "f").files !== undefined &&
            // Similarly, if any of our dependencies are uncacheable, then we're
            // uncacheable too, because that dependency could also have an effect on
            // our output.
            allDependenciesAreCacheable);
    return {
        cacheable,
        platform: process.platform,
        arch: process.arch,
        nodeVersion: process.version,
        command: __classPrivateFieldGet(this, _ScriptExecution_script, "f").command,
        clean: __classPrivateFieldGet(this, _ScriptExecution_script, "f").clean,
        files: Object.fromEntries(fileHashes.sort(([aFile], [bFile]) => aFile.localeCompare(bFile))),
        output: __classPrivateFieldGet(this, _ScriptExecution_script, "f").output ?? [],
        dependencies: Object.fromEntries(filteredDependencyStates.sort(([aRef], [bRef]) => aRef.localeCompare(bRef))),
    };
}, _ScriptExecution_dataDir_get = function _ScriptExecution_dataDir_get() {
    return getScriptDataDir(__classPrivateFieldGet(this, _ScriptExecution_script, "f"));
}, _ScriptExecution_statePath_get = function _ScriptExecution_statePath_get() {
    return pathlib.join(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_dataDir_get), 'state');
}, _ScriptExecution_stdoutReplayPath_get = function _ScriptExecution_stdoutReplayPath_get() {
    return pathlib.join(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_dataDir_get), 'stdout');
}, _ScriptExecution_stderrReplayPath_get = function _ScriptExecution_stderrReplayPath_get() {
    return pathlib.join(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_dataDir_get), 'stderr');
}, _ScriptExecution_readPreviousState = 
/**
 * Read this script's ".wireit/<hex-script-name>/state" file.
 */
async function _ScriptExecution_readPreviousState() {
    try {
        return (await fs.readFile(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_statePath_get), 'utf8'));
    }
    catch (error) {
        if (error.code === 'ENOENT') {
            return undefined;
        }
        throw error;
    }
}, _ScriptExecution_writeStateFile = 
/**
 * Write this script's ".wireit/<hex-script-name>/state" file.
 */
async function _ScriptExecution_writeStateFile(stateStr) {
    await fs.mkdir(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_dataDir_get), { recursive: true });
    await fs.writeFile(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_statePath_get), stateStr, 'utf8');
}, _ScriptExecution_prepareDataDir = 
/**
 * Delete all state for this script from the previous run, and ensure the data
 * directory is created.
 */
async function _ScriptExecution_prepareDataDir() {
    await Promise.all([
        fs.rm(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_statePath_get), { force: true }),
        fs.rm(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stdoutReplayPath_get), { force: true }),
        fs.rm(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stderrReplayPath_get), { force: true }),
        fs.mkdir(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_dataDir_get), { recursive: true }),
    ]);
}, _ScriptExecution_cleanOutput = 
/**
 * Delete all files matched by this script's "output" glob patterns.
 */
async function _ScriptExecution_cleanOutput() {
    if (__classPrivateFieldGet(this, _ScriptExecution_script, "f").output === undefined) {
        return;
    }
    const absFiles = await glob(__classPrivateFieldGet(this, _ScriptExecution_script, "f").output, {
        cwd: __classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir,
        absolute: true,
        followSymlinks: false,
        includeDirectories: true,
        expandDirectories: true,
    });
    if (absFiles.length === 0) {
        return;
    }
    const insidePackagePrefix = __classPrivateFieldGet(this, _ScriptExecution_script, "f").packageDir + pathlib.sep;
    for (const absFile of absFiles) {
        // TODO(aomarks) It would be better to do this in the Analyzer by looking
        // at the output glob patterns, so that we catch errors earlier and can
        // provide a more useful message, but we need to be certain that we are
        // parsing glob patterns correctly (e.g. negations and other syntax make
        // it slightly tricky to detect).
        if (!absFile.path.startsWith(insidePackagePrefix)) {
            throw new WireitError({
                script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                type: 'failure',
                reason: 'invalid-config-syntax',
                message: `refusing to delete output file outside of package: ${absFile.path}`,
            });
        }
    }
    await deleteEntries(absFiles);
}, _ScriptExecution_replayStdoutIfPresent = 
/**
 * Write this script's stdout replay to stdout if it exists, otherwise do
 * nothing.
 */
async function _ScriptExecution_replayStdoutIfPresent() {
    try {
        for await (const chunk of createReadStream(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stdoutReplayPath_get))) {
            __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
                script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                type: 'output',
                stream: 'stdout',
                data: chunk,
            });
        }
    }
    catch (error) {
        if (error.code === 'ENOENT') {
            // There is no saved replay.
            return;
        }
    }
}, _ScriptExecution_replayStderrIfPresent = 
/**
 * Write this script's stderr replay to stderr if it exists, otherwise do
 * nothing.
 */
async function _ScriptExecution_replayStderrIfPresent() {
    try {
        for await (const chunk of createReadStream(__classPrivateFieldGet(this, _ScriptExecution_instances, "a", _ScriptExecution_stderrReplayPath_get))) {
            __classPrivateFieldGet(this, _ScriptExecution_logger, "f").log({
                script: __classPrivateFieldGet(this, _ScriptExecution_script, "f"),
                type: 'output',
                stream: 'stderr',
                data: chunk,
            });
        }
    }
    catch (error) {
        if (error.code === 'ENOENT') {
            // There is no saved replay.
            return;
        }
        throw error;
    }
};
/**
 * Close the given write stream and resolve or reject the returned promise when
 * completed or failed.
 */
const closeWriteStream = (stream) => {
    return new Promise((resolve, reject) => {
        stream.close((error) => {
            if (error) {
                reject(error);
            }
            else {
                resolve();
            }
        });
    });
};
//# sourceMappingURL=executor.js.map