/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
/**
 * Ponyfill for ES2022 AggregateError.
 *
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/AggregateError
 * https://tc39.es/ecma262/multipage/fundamental-objects.html#sec-aggregate-error-objects
 */
export class AggregateError {
    constructor(errors, message = '') {
        Object.defineProperty(this, "errors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "message", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.errors = [...errors];
        this.message = message;
    }
}
//# sourceMappingURL=aggregate-error.js.map