/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _CachingPackageJsonReader_cache;
import * as pathlib from 'path';
import * as fs from 'fs/promises';
/**
 * Reads package.json files and caches them.
 */
export class CachingPackageJsonReader {
    constructor() {
        _CachingPackageJsonReader_cache.set(this, new Map());
    }
    async read(packageDir) {
        let packageJson = __classPrivateFieldGet(this, _CachingPackageJsonReader_cache, "f").get(packageDir);
        if (packageJson === undefined) {
            const packageJsonPath = pathlib.resolve(packageDir, 'package.json');
            let packageJsonStr;
            try {
                packageJsonStr = await fs.readFile(packageJsonPath, 'utf8');
            }
            catch (error) {
                if (error.code === 'ENOENT') {
                    throw new CachingPackageJsonReaderError('missing-package-json');
                }
                throw error;
            }
            try {
                packageJson = JSON.parse(packageJsonStr);
            }
            catch (error) {
                throw new CachingPackageJsonReaderError('invalid-package-json');
            }
            __classPrivateFieldGet(this, _CachingPackageJsonReader_cache, "f").set(packageDir, packageJson);
        }
        return packageJson;
    }
}
_CachingPackageJsonReader_cache = new WeakMap();
/**
 * An exception thrown by {@link CachingPackageJsonReader}.
 *
 * Note we don't use {@link WireitError} here because we don't have the full
 * context of the script we're trying to evaluate.
 */
class CachingPackageJsonReaderError extends Error {
    constructor(reason) {
        super(reason);
        Object.defineProperty(this, "reason", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.reason = reason;
    }
}
//# sourceMappingURL=package-json-reader.js.map