/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Watcher_instances, _Watcher_script, _Watcher_logger, _Watcher_workerPool, _Watcher_watchers, _Watcher_cache, _Watcher_abort, _Watcher_executing, _Watcher_stale, _Watcher_aborted_get, _Watcher_update, _Watcher_watch, _Watcher_analyzeAndExecute, _Watcher_triageErrors, _Watcher_watchPatterns, _Watcher_fileAddedOrChanged, _Watcher_clearWatchers, _Watcher_getWatchPathGroups;
import chokidar from 'chokidar';
import * as pathlib from 'path';
import { Analyzer } from './analyzer.js';
import { Executor } from './executor.js';
import { Deferred } from './util/deferred.js';
import { scriptReferenceToString } from './script.js';
import { WireitError } from './error.js';
import { AggregateError } from './util/aggregate-error.js';
/**
 * Watches a script for changes in its input files, and in the input files of
 * its transitive dependencies, and executes all affected scripts when they
 * change.
 *
 * Also watches all related package.json files and reloads script configuration
 * when they change.
 *
 * State diagram:
 *
 *              ┌──────────────────────────────────┐
 *              |  ┌──────────────┐                |
 *              ▼  ▼              |                |
 *           ┌───────┐      ┌───────────┐      ┌───────┐
 * START ───►| stale | ───► | executing | ───► | fresh |
 *           └───────┘      └───────────┘      └───────┘
 *               │                │                │
 *               |                ▼                |
 *               |           ┌─────────┐           |
 *               └─────────► | aborted | ◄─────────┘
 *                           └─────────┘
 */
export class Watcher {
    constructor(script, logger, workerPool, cache, abort) {
        _Watcher_instances.add(this);
        _Watcher_script.set(this, void 0);
        _Watcher_logger.set(this, void 0);
        _Watcher_workerPool.set(this, void 0);
        _Watcher_watchers.set(this, []);
        _Watcher_cache.set(this, void 0);
        _Watcher_abort.set(this, void 0);
        /** Whether an executor is currently running. */
        _Watcher_executing.set(this, false);
        /** Whether a file has changed since the last time we executed. */
        _Watcher_stale.set(this, true);
        /** Notification that some state has changed. */
        _Watcher_update.set(this, new Deferred());
        /**
         * One of the paths we are watching has been created or modified.
         */
        _Watcher_fileAddedOrChanged.set(this, () => {
            // TODO(aomarks) Cache package JSONS, globs, and hashes.
            __classPrivateFieldSet(this, _Watcher_stale, true, "f");
            __classPrivateFieldGet(this, _Watcher_update, "f").resolve();
        });
        __classPrivateFieldSet(this, _Watcher_script, script, "f");
        __classPrivateFieldSet(this, _Watcher_logger, logger, "f");
        __classPrivateFieldSet(this, _Watcher_workerPool, workerPool, "f");
        __classPrivateFieldSet(this, _Watcher_abort, abort, "f");
        __classPrivateFieldSet(this, _Watcher_cache, cache, "f");
        if (!__classPrivateFieldGet(this, _Watcher_instances, "a", _Watcher_aborted_get)) {
            // eslint-disable-next-line @typescript-eslint/no-floating-promises
            abort.promise.then(() => {
                // TODO(aomarks) Aborting should also cause the analyzer and executors to
                // stop if they are running. Currently we only stop after the current
                // build entirely finishes.
                __classPrivateFieldGet(this, _Watcher_update, "f").resolve();
            });
        }
    }
    /**
     * Execute the script, and continue executing it every time a file related to
     * the configured script changes.
     *
     * @returns When the abort promise resolves.
     * @throws If an unexpected error occurs during analysis or execution.
     */
    static watch(script, logger, workerPool, cache, abort) {
        var _a;
        return __classPrivateFieldGet((_a = new Watcher(script, logger, workerPool, cache, abort)), _Watcher_instances, "m", _Watcher_watch).call(_a);
    }
}
_Watcher_script = new WeakMap(), _Watcher_logger = new WeakMap(), _Watcher_workerPool = new WeakMap(), _Watcher_watchers = new WeakMap(), _Watcher_cache = new WeakMap(), _Watcher_abort = new WeakMap(), _Watcher_executing = new WeakMap(), _Watcher_stale = new WeakMap(), _Watcher_update = new WeakMap(), _Watcher_fileAddedOrChanged = new WeakMap(), _Watcher_instances = new WeakSet(), _Watcher_aborted_get = function _Watcher_aborted_get() {
    return __classPrivateFieldGet(this, _Watcher_abort, "f").settled;
}, _Watcher_watch = 
/**
 * The main watch loop.
 */
async function _Watcher_watch() {
    try {
        while (!__classPrivateFieldGet(this, _Watcher_instances, "a", _Watcher_aborted_get)) {
            if (__classPrivateFieldGet(this, _Watcher_stale, "f") && !__classPrivateFieldGet(this, _Watcher_executing, "f")) {
                await __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_analyzeAndExecute).call(this);
            }
            await __classPrivateFieldGet(this, _Watcher_update, "f").promise;
            __classPrivateFieldSet(this, _Watcher_update, new Deferred(), "f");
        }
    }
    finally {
        // It's important to close all chokidar watchers, because they will
        // prevent the Node program from ever exiting as long as they are active.
        await __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_clearWatchers).call(this);
    }
}, _Watcher_analyzeAndExecute = 
/**
 * Perform an analysis and execution.
 */
async function _Watcher_analyzeAndExecute() {
    __classPrivateFieldGet(this, _Watcher_logger, "f").log({
        script: __classPrivateFieldGet(this, _Watcher_script, "f"),
        type: 'info',
        detail: 'watch-run-start',
    });
    // Reset _stale before execution, not after, because a file could change
    // during execution, and we must not clobber that.
    __classPrivateFieldSet(this, _Watcher_stale, false, "f");
    __classPrivateFieldSet(this, _Watcher_executing, true, "f");
    // TODO(aomarks) We only need to reset watchers and re-analyze if a
    // package.json file changed.
    const analyzer = new Analyzer();
    // TODO(aomarks) Add support for recovering from analysis errors. We'll need
    // to track the package.json files that we encountered, and watch them.
    const analysis = await analyzer.analyze(__classPrivateFieldGet(this, _Watcher_script, "f"));
    await __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_clearWatchers).call(this);
    for (const { patterns, cwd } of __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_getWatchPathGroups).call(this, analysis)) {
        __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_watchPatterns).call(this, patterns, cwd);
    }
    try {
        const executor = new Executor(__classPrivateFieldGet(this, _Watcher_logger, "f"), __classPrivateFieldGet(this, _Watcher_workerPool, "f"), __classPrivateFieldGet(this, _Watcher_cache, "f"));
        await executor.execute(analysis);
    }
    catch (error) {
        __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_triageErrors).call(this, error);
    }
    __classPrivateFieldSet(this, _Watcher_executing, false, "f");
    __classPrivateFieldGet(this, _Watcher_update, "f").resolve();
    __classPrivateFieldGet(this, _Watcher_logger, "f").log({
        script: __classPrivateFieldGet(this, _Watcher_script, "f"),
        type: 'info',
        detail: 'watch-run-end',
    });
}, _Watcher_triageErrors = function _Watcher_triageErrors(error) {
    const errors = error instanceof AggregateError ? error.errors : [error];
    const unexpected = [];
    for (const error of errors) {
        if (error instanceof WireitError) {
            __classPrivateFieldGet(this, _Watcher_logger, "f").log(error.event);
        }
        else {
            unexpected.push(error);
        }
    }
    if (unexpected.length > 0) {
        if (unexpected.length === 1) {
            throw unexpected[0];
        }
        throw new AggregateError(unexpected);
    }
}, _Watcher_watchPatterns = function _Watcher_watchPatterns(patterns, cwd) {
    const watcher = chokidar.watch(patterns, {
        cwd,
        // Ignore the initial "add" events emitted when chokidar first discovers
        // each file. We already do an initial run, so these events are just noise
        // that may trigger an unnecessary second run.
        // https://github.com/paulmillr/chokidar#path-filtering
        ignoreInitial: true,
    });
    __classPrivateFieldGet(this, _Watcher_watchers, "f").push(watcher);
    watcher.on('add', __classPrivateFieldGet(this, _Watcher_fileAddedOrChanged, "f"));
    watcher.on('unlink', __classPrivateFieldGet(this, _Watcher_fileAddedOrChanged, "f"));
    watcher.on('change', __classPrivateFieldGet(this, _Watcher_fileAddedOrChanged, "f"));
}, _Watcher_clearWatchers = 
/**
 * Shut down all active file watchers and clear the list.
 */
async function _Watcher_clearWatchers() {
    const watchers = __classPrivateFieldGet(this, _Watcher_watchers, "f").splice(0, __classPrivateFieldGet(this, _Watcher_watchers, "f").length);
    await Promise.all(watchers.map((watcher) => watcher.close()));
}, _Watcher_getWatchPathGroups = function _Watcher_getWatchPathGroups(script) {
    const packageJsons = new Set();
    const groups = [];
    const visited = new Set();
    const visit = (script) => {
        const key = scriptReferenceToString(script);
        if (visited.has(key)) {
            return;
        }
        visited.add(key);
        packageJsons.add(pathlib.join(script.packageDir, 'package.json'));
        if (script.files !== undefined) {
            // TODO(aomarks) We could optimize to create fewer watchers, but we have
            // to be careful to deal with "!"-prefixed negation entries, because
            // those negations only apply to the previous entries **in that specific
            // "files" array**. A simple solution could be that if a "files" array
            // has any "!"-prefixed entry, then it gets its own watcher, otherwise
            // we can group watchers by packageDir.
            groups.push({ patterns: script.files, cwd: script.packageDir });
        }
        for (const dependency of script.dependencies) {
            visit(dependency);
        }
    };
    visit(script);
    groups.push({
        // The package.json group is already resolved to absolute paths, so cwd is
        // arbitrary.
        cwd: process.cwd(),
        patterns: [...packageJsons],
    });
    return groups;
};
//# sourceMappingURL=watcher.js.map